@@ -18,6 +18,7 @@
  */
 
 #include "PaginateModel.h"
+#include <QtMath>
 #include <QDebug>
 
 PaginateModel::PaginateModel(QObject* object)
@@ -25,6 +26,11 @@ PaginateModel::PaginateModel(QObject* object)
     , m_firstItem(0)
     , m_pageSize(10)
     , m_sourceModel(nullptr)
+    , m_hasStaticRowCount(false)
+{
+}
+
+PaginateModel::~PaginateModel()
 {
 }
 
@@ -58,11 +64,11 @@ void PaginateModel::setPageSize(int count)
         if (difference==0) {
             m_pageSize = count;
         } else if(difference>0) {
-            beginInsertRows(QModelIndex(), m_pageSize, m_pageSize+difference);
+            beginInsertRows(QModelIndex(), m_pageSize, m_pageSize+difference-1);
             m_pageSize = count;
             endInsertRows();
         } else {
-            beginRemoveRows(QModelIndex(), m_pageSize+difference, m_pageSize);
+            beginRemoveRows(QModelIndex(), m_pageSize+difference, m_pageSize-1);
             m_pageSize = count;
             endRemoveRows();
         }
@@ -121,7 +127,9 @@ QHash< int, QByteArray > PaginateModel::roleNames() const
 
 int PaginateModel::rowsByPageSize(int size) const
 {
-    return !m_sourceModel ? 0 : qMin(m_sourceModel->rowCount()-m_firstItem, size);
+    return m_hasStaticRowCount ? size
+        : !m_sourceModel ? 0
+        : qMin(m_sourceModel->rowCount()-m_firstItem, size);
 }
 
 int PaginateModel::rowCount(const QModelIndex& parent) const
@@ -159,7 +167,7 @@ void PaginateModel::firstPage()
 
 void PaginateModel::lastPage()
 {
-    setFirstItem(pageCount()*m_pageSize);
+    setFirstItem((pageCount() - 1)*m_pageSize);
 }
 
 void PaginateModel::nextPage()
@@ -181,8 +189,24 @@ int PaginateModel::pageCount() const
 {
     if(!m_sourceModel)
         return 0;
-    int r = (m_sourceModel->rowCount()%m_pageSize == 0) ? 1 : 0;
-    return m_sourceModel->rowCount()/m_pageSize - r;
+    const int r = (m_sourceModel->rowCount()%m_pageSize == 0) ? 1 : 0;
+    return qCeil(float(m_sourceModel->rowCount())/m_pageSize) - r;
+}
+
+bool PaginateModel::hasStaticRowCount() const
+{
+    return m_hasStaticRowCount;
+}
+
+void PaginateModel::setStaticRowCount(bool src)
+{
+    if (src == m_hasStaticRowCount) {
+        return;
+    }
+
+    m_hasStaticRowCount = src;
+    beginResetModel();
+    endResetModel();
 }
 
 //////////////////////////////
@@ -243,9 +267,9 @@ void PaginateModel::_k_sourceColumnsRemoved(const QModelIndex& parent, int start
     endResetModel();
 }
 
-void PaginateModel::_k_sourceDataChanged(const QModelIndex& topLeft, const QModelIndex& bottomRight)
+void PaginateModel::_k_sourceDataChanged(const QModelIndex& topLeft, const QModelIndex& bottomRight, const QVector<int> &roles)
 {
-    if(topLeft.parent().isValid() || bottomRight.row()<m_firstItem || topLeft.row()>m_firstItem+rowCount()) {
+    if(topLeft.parent().isValid() || bottomRight.row()<m_firstItem || topLeft.row()>lastItem()) {
         return;
     }
 
@@ -256,7 +280,7 @@ void PaginateModel::_k_sourceDataChanged(const QModelIndex& topLeft, const QMode
     if(!idxBottom.isValid())
         idxBottom = index(rowCount()-1);
 
-    emit dataChanged(idxTop, idxBottom);
+    emit dataChanged(idxTop, idxBottom, roles);
 }
 
 void PaginateModel::_k_sourceHeaderDataChanged(Qt::Orientation orientation, int first, int last)
@@ -276,18 +300,41 @@ void PaginateModel::_k_sourceModelReset()
     endResetModel();
 }
 
+bool PaginateModel::isIntervalValid(const QModelIndex& parent, int start, int /*end*/) const
+{
+    return !parent.isValid() && start<=lastItem();
+}
+
+bool PaginateModel::canSizeChange() const
+{
+    return !m_hasStaticRowCount && currentPage() == pageCount()-1;
+}
+
 void PaginateModel::_k_sourceRowsAboutToBeInserted(const QModelIndex& parent, int start, int end)
 {
-    if(parent.isValid() || start>m_firstItem+m_pageSize) {
+    if(!isIntervalValid(parent, start, end)) {
         return;
     }
 
-    int insertedCount = end-start-1;
-    if(insertedCount > m_pageSize) {
-        beginResetModel();
-    } else {
-        int newStart = qMax(start-m_firstItem, 0);
+    if(canSizeChange()) {
+        const int insertedCount = end-start;
+        const int newStart = qMax(start-m_firstItem, 0);
         beginInsertRows(QModelIndex(), newStart, newStart+insertedCount);
+    } else {
+        beginResetModel();
+    }
+}
+
+void PaginateModel::_k_sourceRowsInserted(const QModelIndex& parent, int start, int end)
+{
+    if(!isIntervalValid(parent, start, end)) {
+        return;
+    }
+
+    if(canSizeChange()) {
+        endInsertRows();
+    } else {
+        endResetModel();
     }
 }
 
@@ -298,39 +345,10 @@ void PaginateModel::_k_sourceRowsAboutToBeMoved(const QModelIndex& sourceParent,
     Q_UNUSED(sourceEnd)
     Q_UNUSED(destParent)
     Q_UNUSED(dest)
-    //TODO could optimize, unsure if it makes sense
+    //NOTE could optimize, unsure if it makes sense
     beginResetModel();
 }
 
-void PaginateModel::_k_sourceRowsAboutToBeRemoved(const QModelIndex& parent, int start, int end)
-{
-    if(parent.isValid() || start>m_firstItem+m_pageSize) {
-        return;
-    }
-
-    int removedCount = end-start;
-    if(removedCount > m_pageSize) {
-        beginResetModel();
-    } else {
-        int newStart = qMax(start-m_firstItem, 0);
-        beginRemoveRows(QModelIndex(), newStart, newStart+removedCount);
-    }
-}
-
-void PaginateModel::_k_sourceRowsInserted(const QModelIndex& parent, int start, int end)
-{
-    if(parent.isValid() || start>m_firstItem+m_pageSize) {
-        return;
-    }
-
-    int insertedCount = end-start;
-    if(insertedCount > m_pageSize) {
-        endResetModel();
-    } else {
-        endInsertRows();
-    }
-}
-
 void PaginateModel::_k_sourceRowsMoved(const QModelIndex& sourceParent, int sourceStart, int sourceEnd, const QModelIndex& destParent, int dest)
 {
     Q_UNUSED(sourceParent)
@@ -341,16 +359,35 @@ void PaginateModel::_k_sourceRowsMoved(const QModelIndex& sourceParent, int sour
     endResetModel();
 }
 
-void PaginateModel::_k_sourceRowsRemoved(const QModelIndex& parent, int start, int end)
+void PaginateModel::_k_sourceRowsAboutToBeRemoved(const QModelIndex& parent, int start, int end)
 {
-    if(parent.isValid() || start>m_firstItem+m_pageSize) {
+    if(!isIntervalValid(parent, start, end)) {
         return;
     }
 
-    int removedCount = end-start;
-    if(removedCount > m_pageSize) {
-        beginResetModel();
+    if(canSizeChange()) {
+        const int removedCount = end-start;
+        const int newStart = qMax(start-m_firstItem, 0);
+        beginRemoveRows(QModelIndex(), newStart, newStart+removedCount);
     } else {
-        endRemoveRows();
+        beginResetModel();
     }
 }
+
+void PaginateModel::_k_sourceRowsRemoved(const QModelIndex& parent, int start, int end)
+{
+    if(!isIntervalValid(parent, start, end)) {
+        return;
+    }
+
+    if(canSizeChange()) {
+        endRemoveRows();
+    } else {
+        beginResetModel();
+    }
+}
+
+int PaginateModel::lastItem() const
+{
+    return m_firstItem + rowCount();
+}
